@@ -11,7 +11,7 @@ import logging
 import random
 import string
 import threading
-from typing import Any, Callable, Dict, Coroutine, Optional
+from typing import Any, Callable, Dict, Coroutine, List, Optional
 
 from hyperion import const
 
@@ -29,16 +29,20 @@ class HyperionClientTanNotAvailable(HyperionError):
 class HyperionClientState:
     """Class representing the Hyperion client state."""
 
-    def __init__(self, state={}):
+    def __init__(self, state: Dict = {}) -> None:
         """Initialize state object."""
-        self._state = state
-        self._dirty = False
+        self._state: Dict = state
+        self._dirty: bool = False
 
     @property
     def dirty(self) -> bool:
         """Return whether or state has been modified."""
         return self._dirty
 
+    @dirty.setter
+    def dirty(self, val: bool) -> None:
+        self._dirty = val
+
     def get(self, key: str) -> Any:
         """Retrieve a state element."""
         return self._state.get(key)
@@ -66,14 +70,15 @@ class HyperionClient:
     def __init__(
         self,
         host: str,
-        port: int = const.DEFAULT_PORT,
+        port: int = const.DEFAULT_PORT_JSON,
         default_callback: Optional[Callable] = None,
         callbacks: Optional[Dict] = None,
         token: Optional[str] = None,
         instance: int = const.DEFAULT_INSTANCE,
         origin: str = const.DEFAULT_ORIGIN,
-        timeout_secs: int = const.DEFAULT_TIMEOUT_SECS,
+        timeout_secs: float = const.DEFAULT_TIMEOUT_SECS,
         retry_secs=const.DEFAULT_CONNECTION_RETRY_DELAY_SECS,
+        raw_connection=False,
     ) -> None:
         """Initialize client."""
         _LOGGER.debug("HyperionClient initiated with: (%s:%i)", host, port)
@@ -88,6 +93,7 @@ class HyperionClient:
         self._origin = origin
         self._timeout_secs = timeout_secs
         self._retry_secs = retry_secs
+        self._raw_connection = raw_connection
 
         self._serverinfo: Optional[Dict] = None
 
@@ -112,6 +118,15 @@ class HyperionClient:
             }
         )
 
+    async def __aenter__(self) -> Optional["HyperionClient"]:
+        """Enter context manager and connect the client."""
+        result = await self.async_client_connect()
+        return self if result else None
+
+    async def __aexit__(self, exc_type, exc, tb):
+        """Leave context manager and disconnect the client."""
+        await self.async_client_disconnect()
+
     async def _client_state_reset(self) -> None:
         self._client_state.update(
             {
@@ -167,41 +182,50 @@ class HyperionClient:
     async def async_client_connect(self) -> bool:
         """Connect to the Hyperion server."""
 
+        if self._writer:
+            return True
+
         future_streams = asyncio.open_connection(self._host, self._port)
         try:
             self._reader, self._writer = await asyncio.wait_for(
                 future_streams, timeout=self._timeout_secs
             )
         except (asyncio.TimeoutError, ConnectionError, OSError) as exc:
-            _LOGGER.debug("Could not connect to (%s): %s", self.id, repr(exc))
+            _LOGGER.debug("Could not connect to (%s): %s", self._host_port, repr(exc))
             return False
 
-        _LOGGER.info("Connected to Hyperion server: %s", self.id)
+        _LOGGER.info("Connected to Hyperion server: %s", self._host_port)
 
         # Start the receive task to process inbound data from the server.
         await self._await_or_stop_task(self._receive_task, stop_task=True)
         self._receive_task = asyncio.create_task(self._receive_task_loop())
 
         await self._client_state_reset()
-        self._client_state.set(const.KEY_CONNECTED, True)
-        self._call_client_state_callback_if_necessary()
+        self._client_state.update(
+            {const.KEY_CONNECTED: True, const.KEY_INSTANCE: const.DEFAULT_INSTANCE}
+        )
+        await self._call_client_state_callback_if_necessary()
 
-        if (
-            not self._client_state.get(const.KEY_LOGGED_IN)
-            and not await self._async_client_login()
-        ):
-            return False
+        if not self._raw_connection:
+            if (
+                not self._client_state.get(const.KEY_LOGGED_IN)
+                and not await self._async_client_login()
+            ):
+                await self.async_client_disconnect()
+                return False
 
-        if (
-            not self._client_state.get(const.KEY_INSTANCE)
-            and not await self._async_client_select_instance()
-        ):
-            return False
+            if (
+                not self._client_state.get(const.KEY_INSTANCE)
+                and not await self._async_client_select_instance()
+            ):
+                await self.async_client_disconnect()
+                return False
 
-        if not self._client_state.get(
-            const.KEY_LOADED_STATE
-        ) and not ServerInfoResponseOK(await self.async_get_serverinfo()):
-            return False
+            if not self._client_state.get(
+                const.KEY_LOADED_STATE
+            ) and not ServerInfoResponseOK(await self.async_get_serverinfo()):
+                await self.async_client_disconnect()
+                return False
 
         # Start the maintenance task if it does not already exist.
         if not self._maintenance_task:
@@ -209,7 +233,7 @@ class HyperionClient:
 
         return True
 
-    def _call_client_state_callback_if_necessary(self):
+    async def _call_client_state_callback_if_necessary(self):
         """Call the client state callbacks if state has changed."""
         if not self._client_state.dirty:
             return
@@ -219,13 +243,14 @@ class HyperionClient:
                 const.KEY_COMMAND: f"{const.KEY_CLIENT}-{const.KEY_UPDATE}",
             },
         )
-        self._call_callbacks(str(data[const.KEY_COMMAND]), data)
+        await self._call_callbacks(str(data[const.KEY_COMMAND]), data)
+        self._client_state.dirty = False
 
     async def _async_client_login(self) -> bool:
         """Log the client in if a token is provided."""
         if self._token is None:
             self._client_state.set(const.KEY_LOGGED_IN, True)
-            self._call_client_state_callback_if_necessary()
+            await self._call_client_state_callback_if_necessary()
             return True
         return bool(LoginResponseOK(await self.async_login(token=self._token)))
 
@@ -234,9 +259,9 @@ class HyperionClient:
         if (
             self._client_state.get(const.KEY_INSTANCE) is None
             and self._target_instance == const.DEFAULT_INSTANCE
-        ):
+        ) or self._client_state.get(const.KEY_INSTANCE) == self._target_instance:
             self._client_state.set(const.KEY_INSTANCE, self._target_instance)
-            self._call_client_state_callback_if_necessary()
+            await self._call_client_state_callback_if_necessary()
             return True
 
         resp_json = await self.async_switch_instance(instance=self._target_instance)
@@ -267,13 +292,13 @@ class HyperionClient:
         except ConnectionError as exc:
             _LOGGER.warning(
                 "Could not close connection cleanly for Hyperion (%s): %s",
-                self.id,
+                self._host_port,
                 repr(exc),
             )
             error = True
 
         await self._client_state_reset()
-        self._call_client_state_callback_if_necessary()
+        await self._call_client_state_callback_if_necessary()
 
         # Tell the maintenance loop it may need to reconnect.
         self._maintenance_event.set()
@@ -288,7 +313,7 @@ class HyperionClient:
         if not self._writer:
             return False
 
-        _LOGGER.debug("Send to server (%s): %s", self.id, request)
+        _LOGGER.debug("Send to server (%s): %s", self._host_port, request)
         output = json.dumps(request, sort_keys=True).encode("UTF-8") + b"\n"
         try:
             self._writer.write(output)
@@ -296,29 +321,31 @@ class HyperionClient:
         except ConnectionError as exc:
             _LOGGER.warning(
                 "Could not write data for Hyperion (%s): %s",
-                self.id,
+                self._host_port,
                 repr(exc),
             )
             return False
         return True
 
-    async def _async_safely_read_command(self, timeout: bool = True) -> Optional[Dict]:
+    async def _async_safely_read_command(
+        self, use_timeout: bool = True
+    ) -> Optional[Dict]:
         """Safely read a command from the stream."""
         if not self._reader:
             return None
 
-        timeout_secs = self._timeout_secs if timeout else None
+        timeout_secs = self._timeout_secs if use_timeout else None
 
         try:
             future_resp = self._reader.readline()
             resp = await asyncio.wait_for(future_resp, timeout=timeout_secs)
         except ConnectionError:
-            _LOGGER.warning("Connection to Hyperion lost (%s) ...", self.id)
+            _LOGGER.warning("Connection to Hyperion lost (%s) ...", self._host_port)
             await self._async_client_disconnect_internal()
             return None
         except asyncio.TimeoutError:
             _LOGGER.warning(
-                "Read from Hyperion timed out (%s), disconnecting ...", self.id
+                "Read from Hyperion timed out (%s), disconnecting ...", self._host_port
             )
             await self._async_client_disconnect_internal()
             return None
@@ -327,18 +354,18 @@ class HyperionClient:
             # If there's no writer, we have disconnected, so skip the error message
             # and additional disconnect call.
             if self._writer:
-                _LOGGER.warning("Connection to Hyperion lost (%s) ...", self.id)
+                _LOGGER.warning("Connection to Hyperion lost (%s) ...", self._host_port)
                 await self._async_client_disconnect_internal()
             return None
 
-        _LOGGER.debug("Read from server (%s): %s", self.id, resp)
+        _LOGGER.debug("Read from server (%s): %s", self._host_port, resp)
 
         try:
             resp_json = json.loads(resp)
         except json.decoder.JSONDecodeError:
             _LOGGER.warning(
                 "Could not decode JSON from Hyperion (%s), skipping...",
-                self.id,
+                self._host_port,
             )
             return None
 
@@ -346,7 +373,7 @@ class HyperionClient:
             _LOGGER.warning(
                 "JSON from Hyperion (%s) did not include expected '%s' "
                 "parameter, skipping...",
-                self.id,
+                self._host_port,
                 const.KEY_COMMAND,
             )
             return None
@@ -361,7 +388,7 @@ class HyperionClient:
                         _LOGGER.info(
                             "Could not estalish valid connection to Hyperion (%s), "
                             "retrying in %i seconds...",
-                            self.id,
+                            self._host_port,
                             self._retry_secs,
                         )
                         await self._async_client_disconnect_internal()
@@ -384,7 +411,8 @@ class HyperionClient:
             # Make sure exceptions are logged (for testing purposes, as this is
             # in a background task).
             _LOGGER.exception(
-                "Exception in Hyperion (%s) background maintenance task", self.id
+                "Exception in Hyperion (%s) background maintenance task",
+                self._host_port,
             )
             raise
 
@@ -423,7 +451,7 @@ class HyperionClient:
             {const.KEY_INSTANCE: instance, const.KEY_LOADED_STATE: False}
         )
         self._update_serverinfo(None)
-        self._call_client_state_callback_if_necessary()
+        await self._call_client_state_callback_if_necessary()
 
         # Wake the maintenance task to load the state (this is called from the
         # receive loop, so it cannot be loaded from here).
@@ -431,7 +459,7 @@ class HyperionClient:
 
     async def _async_receive_once(self) -> bool:
         """Manage the bidirectional connection to the server."""
-        resp_json = await self._async_safely_read_command(timeout=False)
+        resp_json = await self._async_safely_read_command(use_timeout=False)
         if not resp_json:
             return False
         command = resp_json[const.KEY_COMMAND]
@@ -443,11 +471,13 @@ class HyperionClient:
                 _LOGGER.warning(
                     "Authorization failed for Hyperion (%s). "
                     "Check token is valid: %s",
-                    self.id,
+                    self._host_port,
                     resp_json,
                 )
             else:
-                _LOGGER.warning("Failed Hyperion (%s) command: %s", self.id, resp_json)
+                _LOGGER.warning(
+                    "Failed Hyperion (%s) command: %s", self._host_port, resp_json
+                )
         elif (
             command == f"{const.KEY_COMPONENTS}-{const.KEY_UPDATE}"
             and const.KEY_DATA in resp_json
@@ -530,8 +560,8 @@ class HyperionClient:
         elif LoginResponseOK(resp_json):
             self._client_state.set(const.KEY_LOGGED_IN, True)
 
-        self._call_callbacks(command, resp_json)
-        self._call_client_state_callback_if_necessary()
+        await self._call_callbacks(command, resp_json)
+        await self._call_client_state_callback_if_necessary()
         await self._handle_response_for_caller(command, resp_json)
         return True
 
@@ -557,19 +587,22 @@ class HyperionClient:
     # || Helper calls ||
     # ==================
 
-    def _call_callbacks(self, command: str, json: Dict) -> None:
+    async def _call_callbacks(self, command: str, json: Dict) -> None:
         """Call the relevant callbacks for the given command."""
-        if command in self._callbacks:
-            self._callbacks[command](json)
-        elif self._default_callback is not None:
-            self._default_callback(json)
+        callback = self._callbacks.get(command, self._default_callback)
+        if callback:
+            if inspect.iscoroutinefunction(callback):
+                await callback(json)
+            else:
+                callback(json)
 
     @property
-    def id(self) -> str:
-        """Return an ID representing this Hyperion client."""
-        return "%s:%i-%i" % (self._host, self._port, self._target_instance)
+    def _host_port(self):
+        """Return a host:port string for this server."""
+        return "%s:%i" % (self._host, self._port)
 
     def _set_data(self, data: Dict, hard: Dict = None, soft: Dict = None) -> Dict:
+        """Override the data in the dictionary selectively."""
         output = soft or {}
         output.update(data)
         output.update(hard or {})
@@ -588,7 +621,7 @@ class HyperionClient:
             if tan in self._tan_responses:
                 raise HyperionClientTanNotAvailable(
                     "Requested tan '%i' is not available in Hyperion client (%s)"
-                    % (tan, self.id)
+                    % (tan, self._host_port)
                 )
             self._tan_responses[tan] = None
             return tan
@@ -599,16 +632,18 @@ class HyperionClient:
             if tan in self._tan_responses:
                 del self._tan_responses[tan]
 
-    async def _wait_for_tan_response(self, tan: int) -> Optional[Dict]:
+    async def _wait_for_tan_response(
+        self, tan: int, timeout_secs: float
+    ) -> Optional[Dict]:
         """Wait for a response to arrive."""
         await self._tan_cv.acquire()
         try:
             await asyncio.wait_for(
                 self._tan_cv.wait_for(lambda: self._tan_responses.get(tan) is not None),
-                timeout=self._timeout_secs,
+                timeout=timeout_secs,
             )
             return self._tan_responses[tan]
-        except asyncio.TimeoutError:  # asyncio.CancelledError
+        except asyncio.TimeoutError:
             pass
         except Exception:
             raise
@@ -625,11 +660,38 @@ class HyperionClient:
     class AwaitResponseWrapper:
         """Wrapper an async *send* coroutine and await the response."""
 
-        def __init__(self, coro):
-            """Initialize the wrapper."""
-            self._coro = coro
+        def __init__(self, coro, timeout_secs: float = 0):
+            """Initialize the wrapper.
 
-        async def __call__(self, client, *args, **kwargs):
+            Wait up to timeout_secs for a response. A timeout of 0
+            will use the client default timeout specified in the constructor.
+            A timeout of None will wait forever.
+            """
+            self._coro = coro
+            self._timeout_secs = timeout_secs
+
+        def _extract_timeout_secs(
+            self, client, data: Dict[str, Any]
+        ) -> Optional[float]:
+            """Return the timeout value for a call.
+
+            Modifies input! Removes the timeout key from the inbound data if
+            present so that it is not passed on to the server. If not present,
+            returns the wrapper default specified in the wrapper constructor.
+            """
+            # Timeout values:
+            #    * None: Wait forever (default asyncio.wait_for behavior).
+            #    * 0: Use the object default (self._timeout_secs)
+            #    * >0: Wait that long.
+            if const.KEY_TIMEOUT_SECS in data:
+                timeout_secs = data[const.KEY_TIMEOUT_SECS]
+                del data[const.KEY_TIMEOUT_SECS]
+                return timeout_secs
+            elif self._timeout_secs == 0:
+                return client._timeout_secs
+            return self._timeout_secs
+
+        async def __call__(self, client, *args: List, **kwargs: Dict[str, Any]):
             """Call the wrapper."""
             # The receive task should never be executing a call that uses the
             # AwaitResponseWrapper (as the response is itself handled by the receive
@@ -639,10 +701,11 @@ class HyperionClient:
 
             tan = await client._reserve_tan_slot(kwargs.get(const.KEY_TAN))
             data = client._set_data(kwargs, hard={const.KEY_TAN: tan})
+            timeout_secs = self._extract_timeout_secs(client, data)
 
             response = None
             if await self._coro(client, *args, **data):
-                response = await client._wait_for_tan_response(tan)
+                response = await client._wait_for_tan_response(tan, timeout_secs)
             await client._remove_tan_slot(tan)
             return response
 
@@ -720,21 +783,21 @@ class HyperionClient:
 
         The user will accept/deny the token request on the Web UI.
         """
-        random_token = "".join(
-            random.choice(string.ascii_letters + string.digits) for i in range(0, 5)
-        )
-
         data = self._set_data(
             kwargs,
             hard={
                 const.KEY_COMMAND: const.KEY_AUTHORIZE,
                 const.KEY_SUBCOMMAND: const.KEY_REQUEST_TOKEN,
             },
-            soft={const.KEY_ID: random_token},
+            soft={const.KEY_ID: generate_random_auth_id()},
         )
         return await self._async_send_json(data)
 
-    async_request_token = AwaitResponseWrapper(async_send_request_token)
+    # This call uses a custom (longer) timeout by default, as the user needs to interact
+    # with the Hyperion UI before it will return.
+    async_request_token = AwaitResponseWrapper(
+        async_send_request_token, timeout_secs=const.DEFAULT_REQUEST_TOKEN_TIMEOUT_SECS
+    )
 
     async def async_send_request_token_abort(self, *args: Any, **kwargs: Any) -> bool:
         """Abort a request for an authorization token."""
@@ -961,7 +1024,7 @@ class HyperionClient:
     # =================================================================================
 
     @property
-    def instances(self) -> Dict:
+    def instances(self) -> List:
         """Return instances."""
         return self._get_serverinfo_value(const.KEY_INSTANCE)
 
@@ -1225,6 +1288,33 @@ class HyperionClient:
 
     async_set_videomode = AwaitResponseWrapper(async_send_set_videomode)
 
+    # ==================================================================================
+    # ** Sysinfo **
+    # Full State: https://docs.hyperion-project.org/en/json/ServerInfo.html#system-hyperion
+    # Returns system information from the Hyperion instance.
+    # ==================================================================================
+
+    async def async_send_sysinfo(self, *args: Any, **kwargs: Any) -> bool:
+        """Request the sysinfo."""
+        data = self._set_data(kwargs, hard={const.KEY_COMMAND: const.KEY_SYSINFO})
+        return await self._async_send_json(data)
+
+    async_sysinfo = AwaitResponseWrapper(async_send_sysinfo)
+
+    async def async_id(self) -> Optional[str]:
+        """Return an ID representing this Hyperion server."""
+        sysinfo = await self.async_sysinfo()
+        if ResponseOK(sysinfo):
+            sysinfo_id = (
+                sysinfo.get(const.KEY_INFO, {})
+                .get(const.KEY_HYPERION, {})
+                .get(const.KEY_ID, None)
+            )
+            if not sysinfo_id:
+                return None
+            return sysinfo_id
+        return None
+
 
 class ThreadedHyperionClient(threading.Thread):
     """Hyperion Client that runs in a dedicated thread."""
@@ -1232,7 +1322,7 @@ class ThreadedHyperionClient(threading.Thread):
     def __init__(
         self,
         host: str,
-        port: int = const.DEFAULT_PORT,
+        port: int = const.DEFAULT_PORT_JSON,
         default_callback: Optional[Callable] = None,
         callbacks: Optional[Dict] = None,
         token: Optional[str] = None,
@@ -1332,7 +1422,7 @@ class ResponseOK:
             return False
         if not self._response.get(const.KEY_SUCCESS, False):
             return False
-        if self._response.get(const.KEY_COMMAND) != self._cmd:
+        if self._cmd is not None and self._response.get(const.KEY_COMMAND) != self._cmd:
             return False
         for validator in self._validators:
             if not validator(self._response):
@@ -1372,3 +1462,10 @@ class SwitchInstanceResponseOK(ResponseOK):
                 lambda r: r.get(const.KEY_INFO, {}).get(const.KEY_INSTANCE) is not None
             ],
         )
+
+
+def generate_random_auth_id():
+    """Generate random authenticate ID."""
+    return "".join(
+        random.choice(string.ascii_letters + string.digits) for i in range(0, 5)
+    )
