@@ -20,6 +20,8 @@ KEY_DATA = "data"
 KEY_EFFECT = "effect"
 KEY_EFFECTS = "effects"
 KEY_ENABLED = "enabled"
+KEY_FRIENDLY_NAME = "friendly_name"
+KEY_HYPERION = "hyperion"
 KEY_LED_MAPPING = "imageToLedMapping"
 KEY_LED_MAPPING_TYPE = "imageToLedMappingType"
 KEY_ID = "id"
@@ -60,14 +62,15 @@ KEY_SUBSCRIBE = "subscribe"
 KEY_SUCCESS = "success"
 KEY_SWITCH_TO = "switchTo"
 KEY_STATE = "state"
+KEY_SYSINFO = "sysinfo"
 KEY_TAN = "tan"
+KEY_TIMEOUT_SECS = "timeout_secs"
 KEY_TOKEN = "token"
 KEY_TOKEN_REQUIRED = "tokenRequired"
 KEY_UPDATE = "update"
 KEY_VALUE = "value"
 KEY_VIDEOMODE = "videomode"
 KEY_VISIBLE = "visible"
-
 KEY_VIDEOMODES = ["2D", "3DSBS", "3DTAB"]
 
 # ComponentIDs from: https://docs.hyperion-project.org/en/json/Control.html#components-ids-explained
@@ -93,5 +96,7 @@ KEY_COMPONENTID_EXTERNAL_SOURCES = [
 DEFAULT_INSTANCE = 0
 DEFAULT_CONNECTION_RETRY_DELAY_SECS = 30
 DEFAULT_TIMEOUT_SECS = 5
+DEFAULT_REQUEST_TOKEN_TIMEOUT_SECS = 180
 DEFAULT_ORIGIN = "hyperion-py"
-DEFAULT_PORT = 19444
+DEFAULT_PORT_JSON = 19444
+DEFAULT_PORT_UI = 8090
